#include "stdafx.h"

#include <fstream>
#include <algorithm>

#include "Util.h"
#include "Logger.h"
#include "NgWordList.h"

////////////////////////////////////////////////////////////////////////
// NG [h
////////////////////////////////////////////////////////////////////////

//----------------------------------------------------------------------
// RXgN^
//----------------------------------------------------------------------

NgWord::NgWord() :
	m_wordType(NgwAll),
	m_bRegex(false),
	m_bCase(false)
{
}

//----------------------------------------------------------------------
// RXgN^
//----------------------------------------------------------------------

NgWord::NgWord(const char* ngword, const WordType wdType, const bool bRegex, const bool bCase) :
	m_word(ngword),
	m_wordType(wdType),
	m_bRegex(bRegex),
	m_bCase(bCase)
{
}

//----------------------------------------------------------------------
// }b`邩
//----------------------------------------------------------------------

bool NgWord::IsMatch(const char* subject, const char* fromname)
{
	bool bError;
	std::string errorWhat;

	if (m_wordType == NgwSubject) {
		return Util::IsMatch(subject, m_word.c_str(), m_bRegex, !m_bCase, bError, errorWhat);
	} else if (m_wordType == NgwFromname) {
		return Util::IsMatch(fromname, m_word.c_str(), m_bRegex, !m_bCase, bError, errorWhat);
	} else /*if (m_wordType == NgwAll)*/ {
		bool matchSubject = Util::IsMatch(subject, m_word.c_str(), m_bRegex, !m_bCase, bError, errorWhat);
		if (matchSubject) {
			return true;
		}
		return Util::IsMatch(fromname, m_word.c_str(), m_bRegex, !m_bCase, bError, errorWhat);
	}
}

//----------------------------------------------------------------------
// NG [h
//----------------------------------------------------------------------

std::string& NgWord::GetWord()
{
	return m_word;
}

//----------------------------------------------------------------------
//  NG [h̎
//----------------------------------------------------------------------

NgWord::WordType NgWord::GetWordType()
{
	return m_wordType;
}

char* NgWord::GetWordTypeStr()
{
	switch (m_wordType) {
	case NgwAll:
	default:
		return "All";
	case NgwSubject:
		return "Subject";
	case NgwFromname:
		return "Fromname";
	}
}

char* NgWord::GetWordTypeJStr()
{
	switch (m_wordType) {
	case NgwAll:
	default:
		return "ׂ";
	case NgwSubject:
		return "";
	case NgwFromname:
		return "ol";
	}
}

//----------------------------------------------------------------------
// K\
//----------------------------------------------------------------------

bool NgWord::IsRegex()
{
	return m_bRegex;
}

//----------------------------------------------------------------------
// p召
//----------------------------------------------------------------------

bool NgWord::IsCase()
{
	return m_bCase;
}

////////////////////////////////////////////////////////////////////////
// NG [hꗗ
////////////////////////////////////////////////////////////////////////

//----------------------------------------------------------------------
// RXgN^
//----------------------------------------------------------------------

NgWordList::NgWordList() :
	m_pcsNgWordList(0),
	m_sortKey(SortNgwType),
	m_bAscending(true)
{
}

//----------------------------------------------------------------------
// NG [hɃ}b`̂邩𓾂
//----------------------------------------------------------------------

bool NgWordList::IsMatch(const char* subject, const char* fromname)
{
	CSLock cslock(m_pcsNgWordList);

	for (std::map<std::string, NgWord>::iterator it = m_ngWordList.begin(); it != m_ngWordList.end(); ++it) {
		if (it->second.IsMatch(subject, fromname)) {
			return true;
		}
	}

	return false;
}

//----------------------------------------------------------------------
// NG [h݂邩𓾂
//----------------------------------------------------------------------

bool NgWordList::IsExist(const char* ngword)
{
	CSLock cslock(m_pcsNgWordList);

	std::map<std::string, NgWord>::iterator it = m_ngWordList.find(ngword);
	if (it != m_ngWordList.end()) {
		return true;
	}

	return false;
}

//----------------------------------------------------------------------
// NG [h̏𓾂
//----------------------------------------------------------------------

std::map<std::string, NgWord>& NgWordList::GetNgWordList()
{
	CSLock cslock(m_pcsNgWordList);

	return m_ngWordList;
}

//----------------------------------------------------------------------
// NG [ȟ𓾂
//----------------------------------------------------------------------

int NgWordList::Count()
{
	CSLock cslock(m_pcsNgWordList);

	return m_ngWordList.size();
}

//----------------------------------------------------------------------
// NG [hǉ
//----------------------------------------------------------------------

void NgWordList::Add(const char* ngword, const NgWord::WordType wdType, const bool bRegex, const bool bCase)
{
	CSLock cslock(m_pcsNgWordList);

	NgWord ngw(ngword, wdType, bRegex, bCase);
	m_ngWordList[ngword] = ngw;

	// \[gp
	std::vector<std::string>::iterator it = std::find(m_sortedNgWord.begin(), m_sortedNgWord.end(), ngword);
	if (it == m_sortedNgWord.end()) {
		m_sortedNgWord.push_back(ngword);
	}
}

//----------------------------------------------------------------------
// 폜
//----------------------------------------------------------------------

void NgWordList::Delete(const char* ngword)
{
#define __FUNCNAME__ "NgWordList::Delete()"

	CSLock cslock(m_pcsNgWordList);

	m_ngWordList.erase(ngword);

	// \[gp
	std::vector<std::string>::iterator it = std::find(m_sortedNgWord.begin(), m_sortedNgWord.end(), ngword);
	if (it != m_sortedNgWord.end()) {
		m_sortedNgWord.erase(it);
	}

#undef __FUNCNAME__
}

//----------------------------------------------------------------------
// XV
//----------------------------------------------------------------------

void NgWordList::Update(const char* ngword, const NgWord::WordType wdType, const bool bRegex, const bool bCase)
{
	// std::map Ȃ̂ŁAgAꉞ֐𕪂Ă
	Add(ngword, wdType, bRegex, bCase);
}

//----------------------------------------------------------------------
// ǉ܂͍XV
//----------------------------------------------------------------------

void NgWordList::AddOrUpdate(const char* ngword, const NgWord::WordType wdType, const bool bRegex, const bool bCase)
{
	// std::map Ȃ̂ŁAgAꉞ֐𕪂Ă
	Add(ngword, wdType, bRegex, bCase);
}

//----------------------------------------------------------------------
// SNA
//----------------------------------------------------------------------

void NgWordList::ClearAll()
{
	CSLock cslock(m_pcsNgWordList);

	m_ngWordList.clear();

	// \[gp
	m_sortedNgWord.clear();
}

//----------------------------------------------------------------------
// C|[g
//----------------------------------------------------------------------

void NgWordList::Import(const char* filepath)
{
	CSLock cslock(m_pcsNgWordList);

	const char* separator = "\t";

	m_ngWordList.clear();
	m_sortedNgWord.clear();	// \[gp

	std::ifstream ngwordlistfile;
	ngwordlistfile.open(filepath, std::ios::in);
	if (ngwordlistfile) {
		std::string line;
		while (std::getline(ngwordlistfile, line)) {

			std::string ngwdata = Util::Trim(line);

			std::string ngword = "";
			NgWord::WordType type = NgWord::NgwAll;
			bool bRegex = false;
			bool bCase = false;

			std::vector<std::string> ngwItems = Util::Split(ngwdata, *separator);
			if (0 < ngwItems.size()) {
				ngword = Util::Trim(ngwItems.at(0).c_str());
				ngword = ngword.substr(0, min(NUM_NG_WORD_MAX, ngword.length()));
				ngword = Util::Trim(ngword);
			}
			if (1 < ngwItems.size()) {
				std::string stype = Util::Trim(ngwItems.at(1).c_str());
				if (stype == "All") {
					type = NgWord::NgwAll;
				} else if (stype == "Subject") {
					type = NgWord::NgwSubject;
				} else if (stype == "Fromname") {
					type = NgWord::NgwFromname;
				}
			}
			if (2 < ngwItems.size()) {
				std::string sRegex = Util::Trim(ngwItems.at(2).c_str());
				if (sRegex == "Regex") {
					bRegex = true;
				} else if (sRegex == "Text") {
					bRegex = false;
				}
			}
			if (3 < ngwItems.size()) {
				std::string sCase = Util::Trim(ngwItems.at(3).c_str());
				if (sCase == "Case") {
					bCase = true;
				} else if (sCase == "IgnoreCase") {
					bCase = false;
				}
			}

			if (0 < ngword.length()) {
				Add(ngword.c_str(), type, bRegex, bCase);
			}
		}

		ngwordlistfile.close();
	}
}

//----------------------------------------------------------------------
// GNX|[g
//----------------------------------------------------------------------

void NgWordList::Export(const char* filepath)
{
	CSLock cslock(m_pcsNgWordList);

	std::ofstream ngwordlistfile;
	ngwordlistfile.open(filepath, std::ios::out | std::ios::trunc);
	if (ngwordlistfile) {

		for(std::map<std::string, NgWord>::iterator it = m_ngWordList.begin(); it != m_ngWordList.end(); ++it) {
			ngwordlistfile << it->second.GetWord() << '\t'
				<< it->second.GetWordTypeStr() << '\t'
				<< ((it->second.IsRegex()) ? "Regex" : "Text") << '\t'
				<< ((it->second.IsCase()) ? "Case" : "IgnoreCase")
				<< std::endl;
		}

		ngwordlistfile.close();
	}
}

//----------------------------------------------------------------------
// rp̃IuWFNgݒ肷
//----------------------------------------------------------------------

void NgWordList::SetLockObject(CRITICAL_SECTION* pcsNgWordList)
{
	m_pcsNgWordList = pcsNgWordList;
}

//----------------------------------------------------------------------
// \[gs
//----------------------------------------------------------------------

//
// \[gp
//
class SortPredNgWord {
public:
	// RXgN^
	SortPredNgWord(NgWordList::SortKey sortKey, bool bAscending, std::map<std::string, NgWord>& ngWordList) :
		m_ngWordList(ngWordList)
	{
		m_sortKey = sortKey;
		m_bAscending = bAscending;
	}

	// r֐
	bool operator()(std::string& left, std::string& right) {
		bool result = false;

		std::map<std::string, NgWord>::iterator itLeft = m_ngWordList.find(left);
		std::map<std::string, NgWord>::iterator itRight = m_ngWordList.find(right);

		// Ce[^ end() ̏ꍇ̃`FbN͔Ô߂̃R[hŁA݂̎gł͋NȂ(͂)B
		if (itLeft == m_ngWordList.end() && itRight == m_ngWordList.end()) {
			result = true;
		} else if (itLeft == m_ngWordList.end() && itRight != m_ngWordList.end()) {
			result = true;
		} else if (itLeft != m_ngWordList.end() && itRight == m_ngWordList.end()) {
			result = false;
		} else {

			const int sortKeyNum = 4;
			NgWordList::SortKey sortKeyArr[sortKeyNum];

			switch (m_sortKey) {
			case NgWordList::SortNgwType:
			default:
				sortKeyArr[0] = NgWordList::SortNgwType;
				sortKeyArr[1] = NgWordList::SortNgwWord;
				sortKeyArr[2] = NgWordList::SortNgwRegex;
				sortKeyArr[3] = NgWordList::SortNgwCase;
				break;
			case NgWordList::SortNgwWord:
				sortKeyArr[0] = NgWordList::SortNgwWord;
				sortKeyArr[1] = NgWordList::SortNgwType;
				sortKeyArr[2] = NgWordList::SortNgwRegex;
				sortKeyArr[3] = NgWordList::SortNgwCase;
				break;
			case NgWordList::SortNgwRegex:
				sortKeyArr[0] = NgWordList::SortNgwRegex;
				sortKeyArr[1] = NgWordList::SortNgwWord;
				sortKeyArr[2] = NgWordList::SortNgwType;
				sortKeyArr[3] = NgWordList::SortNgwCase;
				break;
			case NgWordList::SortNgwCase:
				sortKeyArr[0] = NgWordList::SortNgwCase;
				sortKeyArr[1] = NgWordList::SortNgwWord;
				sortKeyArr[2] = NgWordList::SortNgwType;
				sortKeyArr[3] = NgWordList::SortNgwRegex;
				break;
			}

			int i = 0;
			for ( ; i < sortKeyNum; ++i) {

				switch (sortKeyArr[i]) {
				case NgWordList::SortNgwType:
				default:
					if (itLeft->second.GetWordType() < itRight->second.GetWordType()) {
						result = true;
					} else if (itLeft->second.GetWordType() == itRight->second.GetWordType()) {
						continue;
					} else {
						result = false;
					}
					break;

				case NgWordList::SortNgwWord:
					{
						int comprc = itLeft->first.compare(itRight->first);
						if (comprc < 0) {
							result = true;
						} else if (comprc == 0) {
							continue;	// 蓾Ȃ(͂)BÔ
						} else {
							result = false;
						}
					}
					break;

				case NgWordList::SortNgwRegex:
					if (!itLeft->second.IsRegex() && itRight->second.IsRegex()) {
						result = true;
					} else if (itLeft->second.IsRegex() == itRight->second.IsRegex()) {
						continue;
					} else {
						result = false;
					}
					break;

				case NgWordList::SortNgwCase:
					if (!itLeft->second.IsCase() && itRight->second.IsCase()) {
						result = true;
					} else if (itLeft->second.IsCase() == itRight->second.IsCase()) {
						continue;
					} else {
						result = false;
					}
					break;
				}

				break;
			}

			if (sortKeyNum <= i) {
				result = true;
			}
		}

		if (!m_bAscending) {
			result = !result;
		}

		return result;
	}

private:
	// \[gL[
	NgWordList::SortKey m_sortKey;
	// ǂ
	bool m_bAscending;
	// NG [hꗗ
	std::map<std::string, NgWord>& m_ngWordList;
};

//----------------------------------------------------------------------
// \[gs
//----------------------------------------------------------------------

void NgWordList::Sort()
{
	CSLock cslock(m_pcsNgWordList);

	std::sort(m_sortedNgWord.begin(), m_sortedNgWord.end(), SortPredNgWord(m_sortKey, m_bAscending, m_ngWordList));
}

//----------------------------------------------------------------------
// \[gL[擾/ݒ
//----------------------------------------------------------------------

NgWordList::SortKey NgWordList::GetSortKey()
{
	CSLock cslock(m_pcsNgWordList);

	return m_sortKey;
}

void NgWordList::SetSortKey(SortKey sortKey)
{
	CSLock cslock(m_pcsNgWordList);

	m_sortKey = sortKey;
}

//----------------------------------------------------------------------
// ǂ擾/ݒ
//----------------------------------------------------------------------

bool NgWordList::IsAscending()
{
	CSLock cslock(m_pcsNgWordList);

	return m_bAscending;
}

void NgWordList::SetAscending(bool bAscending)
{
	CSLock cslock(m_pcsNgWordList);

	m_bAscending = bAscending;
}

//----------------------------------------------------------------------
// n Ԗڂ NG [hԂ
//----------------------------------------------------------------------

NgWord& NgWordList::GetAt(int index)
{
	CSLock cslock(m_pcsNgWordList);

	if (index < (int) m_sortedNgWord.size()) {
		std::map<std::string, NgWord>::iterator it = m_ngWordList.find(m_sortedNgWord[index]);
		if (it != m_ngWordList.end()) {
			return it->second;
		}
	}

	// ւ͂Ȃ(͂)
	return m_emptyNgw;
}

//----------------------------------------------------------------------
// w NG [h̃CfbNXԂ
//----------------------------------------------------------------------

int NgWordList::GetIndex(const char* ngword)
{
	CSLock cslock(m_pcsNgWordList);

	std::vector<std::string>::iterator it = std::find(m_sortedNgWord.begin(), m_sortedNgWord.end(), ngword);
	if (it == m_sortedNgWord.end()) {
		return -1;
	}

	int index = std::distance(m_sortedNgWord.begin(), it);
	return index;
}

//----------------------------------------------------------------------
// n Ԗڂ NG [h폜
//----------------------------------------------------------------------

bool NgWordList::DeleteAt(int index)
{
	CSLock cslock(m_pcsNgWordList);

	if (index < (int) m_sortedNgWord.size()) {
		std::map<std::string, NgWord>::iterator it = m_ngWordList.find(m_sortedNgWord[index]);
		if (it != m_ngWordList.end()) {
			std::string ngword = it->first;
			Delete(ngword.c_str());
			return true;
		}
	}

	return false;
}

